from pathlib import Path
from typing import List, Optional, Dict, Union, Tuple, Literal, Sequence, Any
import fsspec
import numpy as np
from xarray import DataArray
from dataclasses import asdict, dataclass
import json
from ..io.mrc import mrc_to_dask
from ..io import read
import dask.array as da
import dacite
from xarray_multiscale.metadata.util import SpatialTransform

CONTAINER_TYPES ={'mrc', 'n5', 'precomputed'}
DTYPE_FORMATS = {"uint16": "n5", "uint8": "precomputed", "uint64": "n5"}
CONTENT_TYPES = {"em", "lm", "prediction", "segmentation", "analysis"}
ContainerTypes = Literal['n5', 'precomputed', 'mrc']


@dataclass
class VolumeStorageSpec:
    kvStore: str
    containerType: ContainerTypes
    containerPath: str
    dataPath: str

    def toURI(self):
        return f'{self.kvStore}://{Path(self.containerPath).with_suffix("." + self.containerType).joinpath(self.dataPath)}'

    def __post_init__(self):
        if self.containerType not in CONTAINER_TYPES:
            raise ValueError(
                f"containerType must be one of {CONTAINER_TYPES}"
            )

@dataclass
class ContrastLimits:
    min: float
    max: float

    def __post_init__(self):
        if not self.min <= self.max:
            raise ValueError('min must be less than or equal to max.')


@dataclass
class DisplaySettings:
    contrastLimits: ContrastLimits
    color: str = 'white'
    invertColormap: bool = False

    @classmethod
    def fromDict(cls, d: Dict[str, Any]):
        return dacite.from_dict(cls, d)


@dataclass 
class DatasetView:
    datasetName: str
    name: str
    description: str
    position: Optional[Sequence[float]]
    scale: Optional[float]
    volumeKeys: Sequence[str]
 
    @classmethod
    def fromDict(cls, d: Dict[str, Any]):
        return dacite.from_dict(cls, d)


@dataclass
class MultiscaleSpec:
    reduction: str
    depth: int
    factors: Union[int, Sequence[int]]


@dataclass
class MeshSource:
    path: str
    name: str
    datasetName: str
    format: str


@dataclass
class VolumeSource:
    path: str
    name: str
    datasetName: str
    dataType: str
    dimensions: Sequence[float]
    transform: SpatialTransform
    contentType: str
    containerType: Optional[ContainerTypes]
    displaySettings: DisplaySettings
    description: str = ''
    version: str="0"
    tags: Optional[Sequence[str]] = None

    def __post_init__(self):
        assert self.contentType in CONTENT_TYPES
        assert len(self.version) > 0

    def toDataArray(self):
        if Path(self.path).suffix == ".mrc":
            array = mrc_to_dask(self.path, chunks=(1, -1, -1))
        else:
            r = read(self.path)
            array = da.from_array(r, chunks=r.chunks)
        coords = [
            DataArray(
                self.transform.translate[idx] + np.arange(array.shape[idx]) * self.transform.scale[idx],
                dims=ax,
                attrs= {'units': self.transform.units[idx]}
                )
            for idx, ax in enumerate(self.transform.axes)
        ]
        return DataArray(array, coords=coords, name=self.name)


    @classmethod
    def fromDict(cls, d: Dict[str, Any]):
        return dacite.from_dict(cls, d)



@dataclass
class DatasetIndex:
    name: str
    volumes: Sequence[VolumeSource]
    meshes: Sequence[MeshSource]
    views: Sequence[DatasetView]
    
    @classmethod
    def from_json(cls, fname: Union[str, Path], open_kwargs: dict = {}):
        with fsspec.open(str(fname), mode='rt', **open_kwargs) as fh:
            jblob = json.loads(fh.read())
        return cls(**jblob)
    
    def to_json(self, fname: Union[str, Path], open_kwargs: dict = {}) -> int:
        jblob = json.dumps(asdict(self))
        with fsspec.open(str(fname), mode='wt', **open_kwargs) as fh:
            result = fh.write(jblob)
        return result


@dataclass
class VolumeIngest:
    source: VolumeSource
    multiscaleSpec: MultiscaleSpec
    storageSpec: VolumeStorageSpec
    mutation: Optional[str] = None


@dataclass
class COSEMArrayAttrs:
    name: str
    transform: SpatialTransform

    @classmethod
    def fromDataArray(cls, data: DataArray) -> "COSEMArrayAttrs":
        name = data.name
        if name is not None:
            return cls(str(name), SpatialTransform.fromDataArray((data)))
        else: 
            raise ValueError('DataArray argument must have a valid name')


@dataclass
class OMEScaleAttrs:
    path: str 
    transform: SpatialTransform


@dataclass
class OMEMultiscaleAttrs:
    datasets: Sequence[OMEScaleAttrs]


@dataclass
class COSEMGroupAttrs:
    name: str
    multiscales: Sequence[OMEMultiscaleAttrs]


@dataclass
class N5PixelResolution:
    dimensions: Sequence[float]
    unit: str


@dataclass
class NeuroglancerGroupAttrs:
    # see https://github.com/google/neuroglancer/issues/176#issuecomment-553027775
    axes: Sequence[str]
    units: Sequence[str]
    scales: Sequence[Sequence[int]]
    pixelResolution: N5PixelResolution


@dataclass
class MultiscaleGroupAttrs:
    name: str
    multiscales: Sequence[OMEMultiscaleAttrs]
    axes: Sequence[str]
    units: Sequence[str]
    scales: Sequence[Sequence[int]]
    pixelResolution: N5PixelResolution


def makeN5ArrayAttrs(dimensions: Sequence[float], unit: str) -> Dict[str, N5PixelResolution]:
    return {'pixelResolution': N5PixelResolution(dimensions, unit)}


def makeMultiscaleGroupAttrs(name: str,
                            arrays: Sequence[DataArray], 
                            array_paths: Sequence[str], 
                            axis_order: str="F") -> MultiscaleGroupAttrs:
    
    assert len(arrays) == len(array_paths)
    cosemArrayAttrs = tuple(COSEMArrayAttrs.fromDataArray(a) for a in arrays)
    
    axis_indexer = slice(None)
    # neuroglancer wants the axes reported in fortran order
    if axis_order == "F":
        axis_indexer = slice(-1, None, -1)
        
    axes: Tuple[str] = arrays[0].dims[axis_indexer]
    scales = tuple(tuple(s.scale_factors)[axis_indexer] for s in arrays)
    coords_reordered = tuple(arrays[0].coords[k] for k in axes)
    units = tuple(d.units for d in coords_reordered)

    # we need this for neuroglancer
    pixelResolution = N5PixelResolution(dimensions=cosemArrayAttrs[0].transform.scale[axis_indexer], unit=units[0])
    multiscales = OMEMultiscaleAttrs(datasets=[OMEScaleAttrs(path=ap, transform=attr.transform) for ap, attr in zip(array_paths, cosemArrayAttrs)])

    result = MultiscaleGroupAttrs(name=name,
                                  multiscales=[multiscales], 
                                  axes=axes,
                                  units=units,
                                  scales=scales,
                                  pixelResolution=pixelResolution)
    return result


@dataclass
class CompositeArrayAttrs:
    name: str
    transform: SpatialTransform
    pixelResolution: N5PixelResolution

    @classmethod
    def fromDataArray(cls, data: DataArray):
        cosemAttrs = COSEMArrayAttrs.fromDataArray(data)
        pixelResolution = N5PixelResolution(cosemAttrs.transform.scale[::-1], unit=cosemAttrs.transform.units[0])
        return cls(cosemAttrs.name, cosemAttrs.transform, pixelResolution)


import numpy as np
import dask.array as da


def pad_arrays(arrays, constant_values, stack=True):
    """
    Pad arrays with variable axis sizes. A bounding box is calculated across all the arrays and each sub-array is
    padded to fit within the bounding box. This is a light wrapper around dask.array.pad. If `stack` is True,
    the arrays will be combined into a larger array via da.stack.

    Parameters
    ----------
    arrays : An iterable collection of dask arrays
    constant_values : The value to fill when padding images

    constant_values : A number which specifies the fill value / mode to use when padding.

    stack: boolean that determines whether the result is a single dask array (stack=True) or a list of dask arrays (stack=False).

    Returns padded arrays and a list of paddings.
    -------

    """

    shapes = np.array([a.shape for a in arrays])
    bounds = shapes.max(0)
    pad_extent = [
        list(zip([0] * shapes.shape[1], (bounds - np.array(a.shape)).tolist()))
        for a in arrays
    ]

    # pad elements of the first axis differently
    def padfun(array, pad_width, constant_values):
        return np.stack(
            [
                np.pad(a, pad_width, constant_values=cv)
                for a, cv in zip(array, constant_values)
            ]
        )

    # If all the shapes are identical no padding is needed.
    if np.unique(shapes, axis=0).shape[0] == 1:
        padded = arrays
    else:
        padded = [
            a.map_blocks(
                padfun,
                pad_width=pad_extent[ind][1:],
                constant_values=constant_values,
                chunks=tuple(
                    c + p[1] - p[0] for c, p in zip(a.chunksize, pad_extent[ind])
                ),
                dtype=a.dtype,
            )
            for ind, a in enumerate(arrays)
        ]

    return padded, pad_extent


def arrays_from_delayed(args, shapes=None, dtypes=None):
    """

    Parameters
    ----------
    args: a collection of dask.delayed objects representing lazy-loaded arrays.

    shapes: a collection of tuples specifying the shape of each array in args, or None. if None, the first array will be loaded
        using local computation, and the shape of that arrays will be used for all subsequent arrays.

    dtypes: a collection of strings specifying the datatype of each array in args, or None. If None, the first array will be loaded
        using local computation and the dtype of that array will be used for all subsequent arrays.

    Returns a list of dask arrays.
    -------

    """

    if shapes is None or dtypes is None:
        sample = args[0].compute(scheduler="threads")
        if shapes is None:
            shapes = (sample.shape,) * len(args)
        if dtypes is None:
            dtypes = (sample.dtype,) * len(args)

    assert len(shapes) == len(args) and len(dtypes) == len(args)

    arrays = [
        da.from_delayed(args[ind], shape=shapes[ind], dtype=dtypes[ind])
        for ind in range(len(args))
    ]
    return arrays



from typing import Tuple, Any, Union, Sequence
from pathlib import Path
from dask import delayed
import dask.array as da
from dask.bag import from_sequence
import zarr
import os
from .util import rmtree_parallel
from toolz import concat


def get_arrays(obj: Any) -> Tuple[zarr.core.Array]:
    result = ()
    if isinstance(obj, zarr.core.Array):
        result = (obj,)
    elif isinstance(obj, zarr.hierarchy.Group):
        if len(tuple(obj.arrays())) > 1:
            names, arrays = zip(*obj.arrays())
            result = tuple(concat(map(get_arrays, arrays)))
    return result


def delete_zbranch(branch, compute=True):
    """
    Delete a branch (group or array) from a zarr container
    """
    if isinstance(branch, zarr.hierarchy.Group):
        return delete_zgroup(branch, compute=compute)
    elif isinstance(branch, zarr.core.Array):
        return delete_zarray(branch, compute=compute)
    else:
        raise TypeError(
            f"The first argument to this function my be a zarr group or array, not {type(branch)}"
        )


def delete_zgroup(zgroup, compute=True):
    """
    Delete all arrays in a zarr group
    """
    if not isinstance(zgroup, zarr.hierarchy.Group):
        raise TypeError(
            f"Cannot use the delete_zgroup function on object of type {type(zgroup)}"
        )

    arrays = get_arrays(zgroup)
    to_delete = delayed([delete_zarray(arr, compute=False) for arr in arrays])

    if compute:
        return to_delete.compute()
    else:
        return to_delete


def delete_zarray(zarray, compute=True):
    """
    Delete a zarr array.
    """

    if not isinstance(zarray, zarr.core.Array):
        raise TypeError(
            f"Cannot use the delete_zarray function on object of type {type(zarray)}"
        )

    path = os.path.join(zarray.store.path, zarray.path)
    store = zarray.store
    branch_depth = None
    if isinstance(store, zarr.N5Store) or isinstance(store, zarr.NestedDirectoryStore):
        branch_depth = 1
    elif isinstance(store, zarr.DirectoryStore):
        branch_depth = 0
    else:
        warnings.warn(
            f"Deferring to the zarr-python implementation for deleting store with type {type(store)}"
        )
        return None

    result = rmtree_parallel(path, branch_depth=branch_depth, compute=compute)
    return result


def same_compressor(arr: zarr.Array, compressor) -> bool:
    """

    Determine if the compressor associated with an array is the same as a different compressor.

    arr: A zarr array
    compressor: a Numcodecs compressor, e.g. GZip(-1)
    return: True or False, depending on whether the zarr array's compressor matches the parameters (name, level) of the
    compressor.
    """
    comp = arr.compressor.compressor_config
    return comp["id"] == compressor.codec_id and comp["level"] == compressor.level


def same_array_props(
    arr: zarr.Array, shape: Tuple[int], dtype: str, compressor: Any, chunks: Tuple[int]
) -> bool:
    """

    Determine if a zarr array has properties that match the input properties.

    arr: A zarr array
    shape: A tuple. This will be compared with arr.shape.
    dtype: A numpy dtype. This will be compared with arr.dtype.
    compressor: A numcodecs compressor, e.g. GZip(-1). This will be compared with the compressor of arr.
    chunks: A tuple. This will be compared with arr.chunks
    return: True if all the properties of arr match the kwargs, False otherwise.
    """
    return (
        (arr.shape == shape)
        & (arr.dtype == dtype)
        & same_compressor(arr, compressor)
        & (arr.chunks == chunks)
    )


def zarr_array_from_dask(arr: Any) -> Any:
    """
    Return the zarr array that was used to create a dask array using `da.from_array(zarr_array)`
    """
    keys = tuple(arr.dask.keys())
    return arr.dask[keys[-1]]


def access_zarr(
    dir_path: Union[str, Path], container_path: Union[str, Path], **kwargs
) -> Any:
    if isinstance(dir_path, Path):
        dir_path = str(dir_path)
    if isinstance(container_path, Path):
        dir_path = str(dir_path)

    attrs = {}
    if "attrs" in kwargs:
        attrs = kwargs.pop("attrs")

    # zarr is extremely slow to delete existing directories, so we do it ourselves
    if kwargs.get("mode") == "w":
        tmp_kwargs = kwargs.copy()
        tmp_kwargs["mode"] = "a"
        tmp = zarr.open(dir_path, path=str(container_path), **tmp_kwargs)
        delete_zbranch(tmp)
    array_or_group = zarr.open(dir_path, path=str(container_path), **kwargs)
    if kwargs.get("mode") != "r":
        array_or_group.attrs.update(attrs)
    return array_or_group


def access_n5(
    dir_path: Union[str, Path], container_path: Union[str, Path], **kwargs
) -> Any:
    dir_path = zarr.N5Store(dir_path)
    return access_zarr(dir_path, container_path, **kwargs)


def zarr_to_dask(store_path: str, key: str, chunks: Union[str, Sequence[int]]):
    arr = access_zarr(store_path, key, mode="r")
    if not hasattr(arr, "shape"):
        raise ValueError(f"{store_path}/{key} is not a zarr array")
    darr = da.from_array(arr, chunks=chunks)
    return darr


def n5_to_dask(store_path: str, key: str, chunks: Union[str, Sequence[int]]):
    arr = access_n5(store_path, key, mode="r")
    if not hasattr(arr, "shape"):
        raise ValueError(f"{store_path}/{key} is not an n5 array")
    darr = da.from_array(arr, chunks=chunks)
    return darr
